﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace InputViewer
{
	/// <summary>
	/// キーボードフック クラス
	/// </summary>
	internal class KeyboardHook
	{
		#region 【Win32 定数・構造体・DllImport】
		/// <summary>
		/// const
		/// </summary>
		protected const int WH_KEYBOARD_LL = 0x000D;
		protected const int WM_KEYDOWN = 0x0100;
		protected const int WM_KEYUP = 0x0101;
		protected const int WM_SYSKEYDOWN = 0x0104;
		protected const int WM_SYSKEYUP = 0x0105;
		protected const int WH_MOUSE_LL = 14;

		// System.Runtime.InteropServices  LayoutKind.Sequential
		[StructLayout(LayoutKind.Sequential)]
		public class KBDLLHOOKSTRUCT
		{
			public uint vkCode;
			public uint scanCode;
			public KBDLLHOOKSTRUCTFlags flags;
			public uint time;
			public UIntPtr dwExtraInfo;
		}

		[Flags]
		public enum KBDLLHOOKSTRUCTFlags : uint
		{
			KEYEVENTF_EXTENDEDKEY = 0x0001,
			KEYEVENTF_KEYUP = 0x0002,
			KEYEVENTF_SCANCODE = 0x0008,
			KEYEVENTF_UNICODE = 0x0004,
		}

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr SetWindowsHookEx(int idHook, KeyboardProc lpfn, IntPtr hMod, uint dwThreadId);

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool UnhookWindowsHookEx(IntPtr hhk);

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, IntPtr wParam, IntPtr lParam);

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr GetModuleHandle(string lpModuleName);
		#endregion 【Win32 定数・構造体】

		#region デリゲート定義・宣言】
		/// <summary>
		/// キーボード処理　デリゲート
		/// </summary>
		/// <param name="nCode"></param>
		/// <param name="wParam"></param>
		/// <param name="lParam"></param>
		/// <returns></returns>
		private delegate IntPtr KeyboardProc(int nCode, IntPtr wParam, IntPtr lParam);
		private KeyboardProc proc;

		public delegate void KeyEventHandler(object sender, KeyEventArg e);
		public event KeyEventHandler KeyEvent;
		#endregion デリゲート定義・宣言】

		#region 【メンバ変数】
		/// <summary>
		/// フックID
		/// </summary>
		private IntPtr hookId = IntPtr.Zero;
		#endregion 【メンバ変数】

		#region 【メソッド】キーボードフック
		/// <summary>
		/// キーボードフック
		/// </summary>
		public void Hook()
		{
			if (hookId == IntPtr.Zero)
			{
				proc = HookProcedure;
				using (var curProcess = Process.GetCurrentProcess())
				{
					using (ProcessModule curModule = curProcess.MainModule)
					{
						hookId = SetWindowsHookEx(WH_KEYBOARD_LL, proc, GetModuleHandle(curModule.ModuleName), 0);
					}
				}
			}
		}
		#endregion 【メソッド】キーボードフック

		#region 【メソッド】キーボードフック解放
		/// <summary>
		/// フック解放
		/// </summary>
		public void UnHook()
		{
			UnhookWindowsHookEx(hookId);
			hookId = IntPtr.Zero;
		}
		#endregion 【メソッド】キーボードフック解放

		#region 【メソッド】フックプロシージャ
		/// <summary>
		/// フックプロシージャ
		/// </summary>
		/// <param name="nCode"></param>
		/// <param name="wParam"></param>
		/// <param name="lParam"></param>
		/// <returns></returns>
		public IntPtr HookProcedure(int nCode, IntPtr wParam, IntPtr lParam)
		{
			if (nCode >= 0 && (wParam == (IntPtr)WM_KEYDOWN || wParam == (IntPtr)WM_SYSKEYDOWN))
			{
				var kb = (KBDLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(KBDLLHOOKSTRUCT));
				var vkCode = (int)kb.vkCode;
				OnKeyDownEvent(vkCode);
			}
			else if (nCode >= 0 && (wParam == (IntPtr)WM_KEYUP || wParam == (IntPtr)WM_SYSKEYUP))
			{
				var kb = (KBDLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(KBDLLHOOKSTRUCT));
				var vkCode = (int)kb.vkCode;
				OnKeyUpEvent(vkCode);
			}
			return CallNextHookEx(hookId, nCode, wParam, lParam);
		}
		#endregion 【メソッド】フックプロシージャ

		#region 【メソッド】キーダウンイベント
		/// <summary>
		/// キーダウンイベント
		/// </summary>
		/// <param name="keyCode"></param>
		protected void OnKeyDownEvent(int keyCode)
		{
			// Invoke からメインスレッド処理呼び出し
			KeyEvent?.Invoke(this, new KeyEventArg(keyCode, true));
		}
		#endregion 【メソッド】キーダウンイベント

		#region 【メソッド】キーアップイベント
		/// <summary>
		/// キーアップイベント
		/// </summary>
		/// <param name="keyCode"></param>
		protected void OnKeyUpEvent(int keyCode)
		{
			// Invoke からメインスレッド処理呼び出し
			KeyEvent?.Invoke(this, new KeyEventArg(keyCode, false));
		}
		#endregion 【メソッド】キーアップイベント

	}

	/// <summary>
	/// キーイベント情報保持クラス(引数)
	/// </summary>
	internal class KeyEventArg
	{
		public int keyCode;
		public bool isDown;

		public KeyEventArg(int a_keyCode, bool a_isDown)
		{
			this.keyCode = a_keyCode;
			this.isDown = a_isDown;
		}
	}
}
